﻿/**
NeGuiの例外クラス。

NeGuiはException投げてもErrorは投げない。
※あくまで明示的にだけど。

History:
	$(HIST_VER 1.00β19,
		$(HIST_NEW)
	)
*/
module etc.negui.system.exception;

debug import std.stdio: wl = writefln;
debug(exception) void main() {}

import std.contracts;
//import std.date;
import std.string;

import etc.negui.system.utility.array; // Error: struct etc.negui.system.text.Text no size yet for forward reference
import etc.negui.system.text;
import etc.negui.system.time;

/**
History:
	$(HIST_VER 1.032,
		$(HIST_F NeGui自体にエラーコードいらねーや。)
	)
*/
class NeGuiException: Exception, IText {
	private {
		//d_time  Time;
		DateTime Time;
	}
	this(in Text msg, Throwable next=null, string _file=__FILE__, int _line=__LINE__) {
		super(msg.text8, _file, _line, next);
		//Time = UTCtoLocalTime(getUTCtime());
		Time = new DateTime;//UTCtoLocalTime(getUTCtime());
	}
	/**
	*/
	final const const(DateTime) time() {
		return Time;
	}

	/**
	Note:
		Throwable.toStringと組み合わせて使用するためconstでない。

	History:
		$(HIST_VER 1.032,
			$(HIST_NEW)
		)
	*/
	protected string ExceptionMessage() {
		return format(
			"%s@%s(%s): %s [%s]",
			this.classinfo.name,
			file,
			line,
			msg,
			time
		);
	}
	/// 未実装。
	override const Text toText() {
		return Text.init;
	}
	
	/**
	History:
		$(HIST_VER 1.032,
			$(HIST_P 処理内容変更。)
		)
	*/
	override string toString() {
		auto s=ExceptionMessage();
		
		if(next) {
			s ~= newline ~ next.toString;
		}

		return s;
	}

}
/***/
mixin template MixInNeGuiException() {
	override this(in Text msg, Throwable next=null, string file=__FILE__, int line=__LINE__) {
		super(msg, next, file, line);
	}
}
/***/
void ThrowableChain(ref Throwable e, Throwable NewThrowable)
in {
	assert(NewThrowable);
}
out {
	assert(e);
}
body {
	if(e) {
		e.next = NewThrowable;
		e = e.next;
	} else {
		e = NewThrowable;
	}
}

/***/
struct THROWABLECHAIN {
	private Throwable base, next;
	/**
	*/
	static nothrow THROWABLECHAIN opCall(Throwable Base) {
		THROWABLECHAIN tc;

		tc.next = tc.base = Base;

		return tc;
	}
	/**
	例外連鎖。

	Params:
		Next = 次の例外。
	*/
	void chain(Throwable Next) {
		ThrowableChain(next, Next);
	}
	/**
	処理継続で捕まえる。
	
	Params:
		expression = 評価する値。

		IsNoThrow = expression評価時にEで受けられる例外を捕まえた場合にfalse、
		            expressionがまともに終わればtrueが入る。

		E = 受け取る例外の種類。

	Returns:
		IsNoThrowがtrueの場合にexpressionの結果。
		
	History:
		$(HIST_VER 1.101,
			$(HIST_F Exception or Errorを取れるように。)
			$(HIST_B expressionがlazyじゃないからダダ漏れ。)
			$(HIST_B IsNoThrowが逆。)
		)
	*/
	T collect(E=Throwable, T, string _file=__FILE__, int _line=__LINE__)(lazy T expression, out bool IsNoThrow) {
		T Result;
		/+
		if(auto e=collectException(expression, Result)) {
			e.file = _file;
			e.line = _line;
			chain(e);
			IsNoThrow = true;
		} else {
			IsNoThrow = false;
		}
		
		return Result;
		+/
		try {
			Result = expression;
			IsNoThrow = true;
		} catch(E e) {
			e.file = _file;
			e.line = _line;
			chain(e);
			IsNoThrow = false;
		}
		
		return Result;
	}
	/**
	これが!nullなら例外あり。
	*/
	Throwable opCall() {
		if(base.next) {
			return base;
		} else {
			return null;
		}
	}
}


